from .SelectedBandLayer  import SelectedBandLayer
from .PointsLayer        import PointsLayer
from .RasterLayer        import RasterLayer
from .RasterRGBLayer     import RasterRGBLayer
from .LayerStack         import LayerStack
from .SwipeLayer         import SwipeLayer
from .Layer              import Layer

layer_types = [SelectedBandLayer, PointsLayer, RasterLayer, SwipeLayer, RasterRGBLayer]
